<!-- HL7 Schema Tightener
	 XSLT to add extra validation to HL7 schemas
	 by Rik Smithies, NProgram Ltd.
 	 (Use at own risk, see disclamer at www.nprogram.co.uk/disclaimer.htm, all rights reserved.)
-->

<xsl:stylesheet version="1.0"  xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:xs="http://www.w3.org/2001/XMLSchema">
<xsl:output encoding="UTF-8"/>
<!--xsl:output method="xml"/-->
<xsl:output indent="yes"/>

<!--
	 History
	 v1.71 18/10/05 Small tidy for release
	 v1.7  12/07/05 Added in a refactoring of fix made in another stream by Ravi Natarajan
	 			    to tighten typeCodes even when no default
	 v1.6  19/04/05 Added optional ability to make default values fixed, with exceptions
	 v1.5  13/04/05 Allows multiple CVs per complexType (some have more than one oid'ed code)
	 v1.4  11/04/05 Made structural code check optional.
			 	    Allow specifying which (one) child element has the specified oid
	 v1.3  06/04/05 Now does some CV codeSystem types also, for GP2GP
	 v1.2  09/03/05 Now adds schema number to new type to prevent name clashes with non-unique types
	 v1.1  08/03/05 Added ability to contrain II types to have fixed OIDs, based on data in SchemaTightenerTypesList.xml
	 v1.0  06/01/05 Created. First version changes all 'optional' attributes to 'required'
-->

<xsl:variable name="sgSchemaTightenerTypesList" select="'SchemaTightenerTypesList.xml'"/>
<!-- Optionally suppress tightening of structural codes (mainly for testing the OID
     enforcing code and types list data -->
<xsl:param name="bgNoStructuralCodeTightening" select="false()"/>

<!-- Use bgMakeAllDefaultValuesFixed with care. Setting this 'true' is too restrictive for general use,
     in particular it disallows valid subtypes of 'fixed' values so is not suitable
     for clinical domains (but may be ok for domains such as LR, PDS etc) -->
<xsl:param name="bgMakeAllDefaultValuesFixed" select="false()"/>

<xsl:template match="/">
	<xsl:comment> This is a non-standard "strict" version of a schema, with optional elements made required and II/CV types tightened </xsl:comment>
	<xsl:comment> It is provided "without prejudice" with no guarantee of fitness for any purpose </xsl:comment>
	<xsl:comment><xsl:text> Produced by XSDStrictMaker v1.71,</xsl:text>
	<xsl:value-of select="concat(' Definition file v',document($sgSchemaTightenerTypesList)/dataroot/@version)"/>
	<xsl:text>, Rik Smithies, NProgram Ltd 15/04/2005 </xsl:text></xsl:comment>
	<xsl:apply-templates/>
</xsl:template>

<!-- identity template to copy all non-root elements unless overridden elsewhere -->
<xsl:template match="@*|node()">
	<xsl:copy>
		<xsl:apply-templates select="@*|node()"/>
	</xsl:copy>
</xsl:template>

<!-- extra templates to intercept and modify certain nodes -->

<!-- convert optional structural attributes to be required -->
<xsl:template match="node()[@use='optional' and @default]">
	<xsl:call-template name="tightenAttribute"/>
</xsl:template>

<!-- special case for typeCode which had no default in some MIM 2.3 messages -->
<!-- a refactoring of Ravi's change -->
<xsl:template match="node()[@use='optional' and @name='typeCode']">
	<xsl:call-template name="tightenAttribute"/>
</xsl:template>

<xsl:template name="tightenAttribute">
	<xsl:element name="{name(.)}">
		<xsl:for-each select="@*">
			<xsl:choose>
				<xsl:when test="name(.)='use' and not ($bgNoStructuralCodeTightening=true())">
					<xsl:attribute name="use"><xsl:value-of select="'required'"/></xsl:attribute>
				</xsl:when>
				<!-- It is too restrictive to turn default into fixed, must remove it -->
				<!-- To do : do this optionally for ones known to be fixed  -->
				<xsl:when test="name(.)='default'  and not ($bgNoStructuralCodeTightening=true())">
					<xsl:if test="$bgMakeAllDefaultValuesFixed=true()">
						<xsl:variable name="sTypeName" select="substring-after(../../@name,'.')"/> <!-- todo relies on dot in name -->
						<xsl:if test="count(document($sgSchemaTightenerTypesList)//elementsToIgnoreDefaultsFor/*[@name=$sTypeName])=0">
							<xsl:attribute name="fixed"><xsl:value-of select="."/></xsl:attribute>
						</xsl:if>
					</xsl:if>
				</xsl:when>
				<xsl:otherwise>	<!-- send it out unchanged -->
					<xsl:attribute name="{name(.)}"><xsl:value-of select="."/></xsl:attribute>
				</xsl:otherwise>
			</xsl:choose>
		</xsl:for-each>
		<xsl:apply-templates/>
	</xsl:element>
</xsl:template>

<!-- OID checking code -->
<!-- Works in two steps
     1. finds complexTypes that use II (or CV) within them, and inserts extra code before
     the existing type that has a new type, that is derived from II
 	 but is more strict (eg. adds a fixed value to the "root" attribute).
 	 The new type is given a new name, say "IIPatient"
 	 2. other code processes the body of the complexType found in step 1
 	 and changes references to II to point to the new type called IIPatient, instead of plain II -->

<!-- Step 1 (for II) template to insert new II derived type into XSD, which will be used later on -->
<xsl:template match="xs:complexType[.//@type='II']">

	<xsl:variable name="sBaseType" select="'II'"/>
	<xsl:variable name="sTypeName" select="substring-after(@name,'.')"/>
	<xsl:variable name="sTypePrefix" select="substring-before(@name,'.')"/>

	<!-- To do : would be good to assert that there really is a prefix -->

	<!-- handle tightening of this complexType -->
	<xsl:call-template name="doComplexTypeChildElementsTightening">
		<xsl:with-param name="sTypeName" select="$sTypeName"/>
		<xsl:with-param name="sTypePrefix" select="$sTypePrefix"/>
		<xsl:with-param name="sBaseType" select="$sBaseType"/>
	</xsl:call-template>

</xsl:template>

<!-- this is the core of code to add new stricted types, called by Step 1 of the tightening process -->
<xsl:template name="doComplexTypeChildElementsTightening">

	<xsl:param name="sTypeName"/>
	<xsl:param name="sTypePrefix"/>
	<xsl:param name="sBaseType"/>

	<!-- all we know at this stage is that this is a complexType that has one, or more CV/II types in it (sTypeName)
		 we dont know if any of these are of interest to us -->

	<!-- loop over possibly several named child elements of this complexTypes from our types list -->
	<xsl:for-each select="document($sgSchemaTightenerTypesList)//tighten/*[@name=$sTypeName]">

		<xsl:comment>
			<xsl:value-of select="concat(' XSDStrictMaker : New stricter complex type derived from ',$sBaseType,' ')"/>
		</xsl:comment>

		<xsl:variable name="sOID" select="@oid"/>
		<xs:complexType>
			<xsl:attribute name="name">
				<xsl:value-of select="concat($sBaseType,$sTypeName,'_',@child,'_',$sTypePrefix)"/>
			</xsl:attribute>
			<xs:complexContent>
				<xs:restriction base="{$sBaseType}">
					<xsl:choose>
					<xsl:when test="$sBaseType='II'">
						<xs:attribute name="root" type="uid">
							<xsl:attribute name="fixed">
								<xsl:value-of select="$sOID"/>
							</xsl:attribute>
						</xs:attribute>
						<xs:attribute name="extension" type="st" use="required"/>
					</xsl:when>
					<xsl:when test="$sBaseType='CV'">
						<xs:attribute name="codeSystem" type="uid">
							<xsl:attribute name="fixed">
								<xsl:value-of select="$sOID"/>
							</xsl:attribute>
						</xs:attribute>
					</xsl:when>
					<xsl:otherwise>
						<!-- no other known cases -->
					</xsl:otherwise>
					</xsl:choose>
				</xs:restriction>
			</xs:complexContent>
		</xs:complexType>

	</xsl:for-each>

	<!-- new preamble to existing complexType has been sent out (or possibly nothing was done) -->
	<!-- now send out the complex type itself, with a comment in front -->

	<!-- if this type was not mentioned in the interesting list, try to find a reason why not, otherwise
		 conclude it has been forgotten. Add a comment whichever -->
	<xsl:choose>
		<!-- see if this is one we expected to change -->
		<xsl:when test="count(document($sgSchemaTightenerTypesList)//tighten/*[@name=$sTypeName])=0">
			<!-- this is not specified as one to alter, so send it out unchanged, but with a comment -->
			<!-- add comment for other types not specifically handled -->
			<xsl:apply-templates select="." mode="ignoredTypes">
				<xsl:with-param name="sTypeName" select="$sTypeName"/>
			</xsl:apply-templates>
		</xsl:when>

		<!-- this is one from out list so it must have had a preamble added -->
		<xsl:otherwise>
			<xsl:comment>
				<xsl:value-of select="concat(' XSDStrictMaker : Type uses new stricter complex type derived from ',$sBaseType,' ')"/>
			</xsl:comment>
		</xsl:otherwise>
	</xsl:choose>

	<xsl:copy>
		<xsl:apply-templates select="@*|node()"/>
	</xsl:copy>

</xsl:template>

<!-- This is step 2 of tightening process, inserting new complex type derived from II (or whatever)
	 As it stands this is just an extra unused type, it needs step 1, below to change the
	 complexType to actually use this version of II instead of II itself -->

<!-- Template to use new type derived from II that was inserted just previously -->
<!-- now also works for CV types -->
<xsl:template match="xs:element[@type='II' or @type='CV']">

	<xsl:variable name="sTypeName" select="substring-after(ancestor::xs:complexType[1]/@name,'.')"/>
	<xsl:variable name="sTypePrefix" select="substring-before(ancestor::xs:complexType[1]/@name,'.')"/>

	<xsl:variable name="theCodedType" select="."/>

	<xsl:variable name="sDataType" select="@type"/>

	<xsl:variable name="sNoChildVersion" select="document($sgSchemaTightenerTypesList)//tighten/*[@name=$sTypeName and not(@child)]/@name"/>
	<xsl:variable name="sMatchingChildVersion" select="document($sgSchemaTightenerTypesList)//tighten/*[@name=$sTypeName and @child=$theCodedType/@name]/@child"/>

	<xsl:choose>
		<xsl:when test="string-length($sMatchingChildVersion) != 0 or string-length($sNoChildVersion) != 0">
			<xsl:comment>
				<xsl:value-of select="concat(' ',$sTypeName,' OID tightened [',$sDataType,'] ')"/>
			</xsl:comment>
			<xsl:element name="{name(.)}">
				<xsl:for-each select="@*">
					<xsl:choose>
						<xsl:when test="name(.)='type'">
							<xsl:attribute name="type">
								<xsl:value-of select="concat($sDataType,$sTypeName,'_',$sMatchingChildVersion,'_',$sTypePrefix)"/>
							</xsl:attribute>
						</xsl:when>
						<xsl:otherwise>
							<xsl:attribute name="{name(.)}"><xsl:value-of select="."/></xsl:attribute>
						</xsl:otherwise>
					</xsl:choose>
				</xsl:for-each>
				<xsl:apply-templates/>
			</xsl:element>
		</xsl:when>
		<xsl:otherwise>
			<xsl:copy>
				<xsl:apply-templates select="@*|node()"/>
			</xsl:copy>
		</xsl:otherwise>
	</xsl:choose>

</xsl:template>

<xsl:template match="*" mode="ignoredTypes">
	<xsl:param name="sTypeName"/>

	<!-- this is not specified as one to alter, so send it out unchanged, but with a comment -->
	<xsl:choose>
		<!-- in the exclusions list ? -->
		<xsl:when test="count(document($sgSchemaTightenerTypesList)//ignore/*[@name=$sTypeName])=1">
			<xsl:comment> XSDStrictMaker : ComplexType specifically ignored </xsl:comment>
		</xsl:when>
		<!-- some special cases, that get special comments, to aid in checking that all OIDs are covered -->
		<!-- nb these ad-hoc rules only produce comments,
		 	 it is only the specific inclusions that cause OID tests to go into XSD -->
		<!-- special case of 'Ref's (generally) not needing OIDs -->
		<xsl:when test="substring($sTypeName,(string-length($sTypeName)-string-length('Ref'))+1)='Ref'">
			<xsl:comment> XSDStrictMaker : ComplexType ignored due to ending in '-Ref' </xsl:comment>
		</xsl:when>
		<!-- special case of pair of IIs implying UUID not an OIDed id  -->
		<!-- nb this test looks for any pair of IIs in this type, ignoring possible other single II -->
		<xsl:when test=".//xs:element[@type='II' and @maxOccurs='2']">
			<xsl:comment> XSDStrictMaker : ComplexType ignored due to II pair </xsl:comment>
		</xsl:when>
		<!-- special case of complexType that is an element at top level of schema - in otherwords the message itself -->
		<xsl:when test="/*/xs:element[@name=$sTypeName]">
			<xsl:comment> XSDStrictMaker : ComplexType ignored due to being top level element </xsl:comment>
		</xsl:when>
		<xsl:otherwise>
			<xsl:comment> XSDStrictMaker : ComplexType not changed </xsl:comment>
		</xsl:otherwise>
	</xsl:choose>

</xsl:template>

<!-- CV, CD, CE handling -->
<!-- To do : add "or .//@type='CD' or .//@type='CE'" -->
<xsl:template match="xs:complexType[.//@type='CV']">

	<xsl:variable name="sBaseType" select="'CV'"/>
	<xsl:variable name="sTypeName" select="substring-after(@name,'.')"/>
	<xsl:variable name="sTypePrefix" select="substring-before(@name,'.')"/>

	<!-- handle tightening of this complexType -->
	<xsl:call-template name="doComplexTypeChildElementsTightening">
		<xsl:with-param name="sTypeName" select="$sTypeName"/>
		<xsl:with-param name="sTypePrefix" select="$sTypePrefix"/>
		<xsl:with-param name="sBaseType" select="$sBaseType"/>
	</xsl:call-template>

</xsl:template>

</xsl:stylesheet>